# A SAT Attack on Rota's Basis Conjecture

## Installation

Using [conda](https://docs.conda.io/en/latest/) (or a similar package management system such as miniconda, anaconda, or mamba) one can install clingo (possibly an environment has to be created first):

```
conda install -c potassco/label/dev clingo
```
Install the SAT solver [cadical](https://github.com/arminbiere/cadical/) and the DRAT verification tool [drat-trim](http://github.com/marijnheule/drat-trim).

## The Rota program

In the following, we describe how our SMS framework can be used to verify Rota's basis conjecture for ranks 3, 4, and 5. We enter the directory "rotaEnc" and run
```
cmake . && make
```
to compile C program using gcc. This creates three executables "rota3", "rota4" and "rota5" in the subdirectory "src/". 
To verify Rota's Basis Conjecture for example for rank r=5 and the parameters girth g=5 and k=3 disjoint rainbow bases, we start SMS and clingo to produce an unsatisfiable CNF formula
```
./src/rota5 -o r5g5k3test.extcnf --girth 5 -k 3 
```
The parameters "--girth" and "-k" here are both optional and can be omitted for example to verify the conjecture for ranks 3 or 4. The program creates an extended CNF formula at r5g5t3test.extcnf, which contains an unsatisfiable CNF with additional information: for each clause there is an explaination, why this clause was added. The correctness of the clauses in the extended CNF formula can be verified with the Python script "verify.py".
```
python scripts/verify.py -i r5g5k3test.extcnf -r 5 -g 5 -k 3 
```
Using the Python script "verify2dimacs.py", we remove the explainations from the extended CNF formula and create CNF formuka in the classical DMACS format.
```
python scripts/verify2dimacs.py -i r5g5k3test.extcnf > r5g5k3test.cnf
```
The unsatisfiablity of this CNF instance is then checked by cadical, which also creates a DRAT certificate.
```
cadical -q --unsat r5g5k3test.cnf r5g5k3test.drat
```
Last but not least, the correctness of the DRAT certificate can be verified using the tool DRAT-trim.
```
drat-trim r5g5k3test.cnf r5g5k3test.drat -t 99999999
```
## Usage of the matroid enumeration program
In the following, we describe how our SMS framework can be used to enumerate all non-isomorphic matroids. To compile the C program using gcc, we enter the directory "enumNonIso" and run
```
cmake . && make
```
This creates an executable "enummatroids" in the subdirectory "src/". To enumerate the number of rank r=3 matroids on n=7 elements, run
```
./src/enummatroids -r 3 -n 7
```

## Example output for rank 4
The folder "data" contains the output for solving RBC for rank 4.
We used the following command to generate the outputs.
```
./src/rota4 -o ./data/rota4.extcnf
python scripts/verify2dimacs.py -i ./data/rota4.extcnf > ./data/rota4.cnf
cadical -q --unsat ./data/rota4.cnf ./data/rota4.drat
```

The files can be verified using
```
python scripts/verify.py -i rota.extcnf -r 4
drat-trim ./data/rota4.cnf ./data/rota4.drat -t 99999999
```

