#include "clause.h"

/**
 * Things which must hold
 * p1,p2 only point to unassigned literals OR p1 on a true literal and p2 = -1 OR both -1
 * if ready 2 fire then p1=p2=-1
 * cl.indices[] give the indices for p1 and p2 where to look at
 *
 */

// get new stat of the given clause. // TODO refine to use previous information
void getStats(clause_t cl, truth_value_t *assignment, int *pos1, int *pos2, int *ready2fire, int *trueLit)
{
    *trueLit = -1;
    *ready2fire = 0;

    for (int i = 0; i < cl.n; i++)
    {
        if (assignment[cl.literals[i]] == truth_value_true)
        {
            *trueLit = i;
            return;
        }
    }

    for (*pos1 = 0; *pos1 < cl.n - 1; (*pos1)++)
    {
        if (assignment[cl.literals[*pos1]] == truth_value_false)
            continue;
        break;
    }

    for (*pos2 = cl.n - 1; *pos2 > *pos1; (*pos2)--)
    {
        if (assignment[cl.literals[*pos2]] == truth_value_false)
            continue;
        break;
    }

    *ready2fire = (*pos1 == *pos2);
}

clause_t *addNewClause2(clauses_state_t *state, truth_value_t *assignment, int b1, int b2, int e, int *literals, int size)
{
    // Create new clause and add to lists add clause to state and check update list.
    if (state->clauseList.n + 1 == state->clauseList.maxN)
    {
        state->clauseList.maxN *= 2;
        state->clauseList.clauses = realloc(state->clauseList.clauses, sizeof(clause_t) * state->clauseList.maxN);
        CHECK_OUT_OF_MEMORY(state->clauseList.clauses)
        printf("c\tSize of clauses increased to %d\n", state->clauseList.maxN);
    }

    int index = state->clauseList.n;
    clause_t *clause = &(state->clauseList.clauses[index]);
    clause->b1 = b1; // TODO check if lit and not base indices.
    clause->b2 = b2;
    clause->p1 = -1;
    clause->p2 = -1;
    clause->e = e;
    clause->n = size;
    clause->indexFire = -1;
    clause->satisfied = 0;
    for (int i = 0; i < size; i++)
    {
        clause->indices[i] = -1;
        clause->literals[i] = literals[i];
    }

    assert(size > 1);

    int p1, p2, ready2fire, trueLit;
    getStats(*clause, assignment, &p1, &p2, &ready2fire, &trueLit);
    if (trueLit != -1)
        return NULL; // must not be added
    else if (ready2fire)
    {
        clause->indexFire = addElemToIntList(&state->ready2fire, index);
    }
    else
    {
        assert(p1 != p2);
        assert(p1 != -1);
        assert(p2 != -1);
        clause->p1 = p1;
        clause->p2 = p2;
        // p1 != p2 because otherwise ready to fire
        clause->indices[p1] = addElemToIntList(&state->lit2clauses[clause->literals[p1]], index);
        clause->indices[p2] = addElemToIntList(&state->lit2clauses[clause->literals[p2]], index);
    }

    clause->indexB1 = addElemToIntList(&state->baseLits2clauses[clause->b1], index);
    clause->indexB2 = addElemToIntList(&state->baseLits2clauses[clause->b2], index);

    state->clauseList.n++; // Only added at this point

    if (clause->p1 != -1 && clause->indices[clause->p1] == -1)
    {
        printf("%d %d\n", clause->p1, clause->indices[clause->p1]);
        EXIT_UNWANTED_STATE
    }
    return clause;
}

void removeP1(clauses_state_t *state, clause_t *clause)
{
    int i = clause->p1;
    assert(i != -1);
    int index2Remove = clause->indices[i];
    int_list_t *list = &(state->lit2clauses[clause->literals[i]]);
    list->n--;
    if (index2Remove != list->n)
    {
        list->elems[index2Remove] = list->elems[list->n];
        clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);

        for (int j = 0; j < cl->n; j++) // search for same literal
        {
            if (cl->literals[j] == clause->literals[i])
            {
                cl->indices[j] = index2Remove;
                break;
            }
        }
    }
    clause->indices[clause->p1] = -1;
    clause->p1 = -1;
}

void removeP2(clauses_state_t *state, clause_t *clause)
{
    int i = clause->p2;
    assert(i != -1);
    int index2Remove = clause->indices[i];
    int_list_t *list = &(state->lit2clauses[clause->literals[i]]);
    list->n--;
    if (index2Remove != list->n)
    {
        list->elems[index2Remove] = list->elems[list->n];
        clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);

        for (int j = 0; j < cl->n; j++) // search for same literal
        {
            if (cl->literals[j] == clause->literals[i])
            {
                cl->indices[j] = index2Remove;
                break;
            }
        }
    }
    clause->indices[clause->p2] = -1;
    clause->p2 = -1;
}

void removeP1andP2(clauses_state_t *state, clause_t *clause)
{
    if (clause->p1 != -1)
        removeP1(state, clause);
    if (clause->p2 != -1)
        removeP2(state, clause);

    if (clause->p1 != -1 && clause->indices[clause->p1] == -1)
    {
        printf("%d %d\n", clause->p1, clause->indices[clause->p1]);
        EXIT_UNWANTED_STATE
    }
}

void removeReady2fire(clauses_state_t *state, clause_t *clause)
{
    int index2Remove = clause->indexFire;
    if (index2Remove == -1)
        return;
    int_list_t *list = &(state->ready2fire);
    list->n--;
    if (index2Remove != list->n)
    {
        list->elems[index2Remove] = list->elems[list->n];
        clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
        cl->indexFire = index2Remove;
    }
    clause->indexFire = -1;

    if (clause->p1 != -1 && clause->indices[clause->p1] == -1)
    {
        printf("%d %d\n", clause->p1, clause->indices[clause->p1]);
        EXIT_UNWANTED_STATE
    }
}

void updateClause(clauses_state_t *state, clause_t *clause, truth_value_t *assignment, int clauseIndex)
{
    int p1, p2, ready2fire, trueLitIndex;
    getStats(*clause, assignment, &p1, &p2, &ready2fire, &trueLitIndex);
    if (trueLitIndex != -1)
    {
        removeReady2fire(state, clause);
        removeP1andP2(state, clause); // TODO wastefull to remove and add again if little has changed
        clause->p1 = trueLitIndex;
        clause->indices[clause->p1] = addElemToIntList(&state->lit2clauses[clause->literals[clause->p1]], clauseIndex);
        clause->satisfied = 1;
    }
    else if (ready2fire)
    {
        if (clause->indexFire != -1)
            return;
        removeReady2fire(state, clause);
        removeP1andP2(state, clause); // TODO wastefull to remove and add again if little has changed
        clause->satisfied = 0;
        clause->indexFire = addElemToIntList(&state->ready2fire, clauseIndex);
    }
    else
    {
        removeReady2fire(state, clause);
        removeP1andP2(state, clause); // TODO wastefull to remove and add again if little has changed
        clause->satisfied = 0;

        clause->p1 = p1;
        clause->p2 = p2;
        // p1 != p2 because otherwise ready to fire
        clause->indices[clause->p1] = addElemToIntList(&state->lit2clauses[clause->literals[p1]], clauseIndex);
        clause->indices[clause->p2] = addElemToIntList(&state->lit2clauses[clause->literals[p2]], clauseIndex);
    }
}

// Update the clause state after adding the literal lit
void addLiteral2(clauses_state_t *state, lit_t lit, truth_value_t *assignment)
{
    if (lit > 0)
    {
        int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated

        int n = list.n;
        int clauseIndices[n]; // copy so we can change them without any concerns
        memcpy(clauseIndices, list.elems, sizeof(int) * n);

        for (int i = 0; i < n; i++)
        {
            clause_t *clause = &state->clauseList.clauses[clauseIndices[i]];
            clause->satisfied = 1;
            for (int litIndex = 0; litIndex < clause->n; litIndex++)
            {
                if (clause->literals[litIndex] == lit)
                {
                    removeP1andP2(state, clause);
                    clause->p1 = litIndex;
                    clause->indices[clause->p1] = addElemToIntList(&state->lit2clauses[clause->literals[clause->p1]], clauseIndices[i]);
                }
            }
        }
    }
    else
    {
        int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated

        int n = list.n;
        int clauseIndices[n]; // copy so we can change them without any concerns
        memcpy(clauseIndices, list.elems, sizeof(int) * n);

        for (int i = 0; i < n; i++)
        {
            clause_t *clause = &state->clauseList.clauses[clauseIndices[i]];
            updateClause(state, clause, assignment, clauseIndices[i]);
        }
    }
}

// Update the clause state after removing the literal lit
void removeLiteral2(clauses_state_t *state, lit_t lit, truth_value_t *assignment)
{
    int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated

    int n = list.n;
    int clauseIndices[n]; // copy so we can change them without any concerns
    memcpy(clauseIndices, list.elems, sizeof(int) * n);

    for (int i = 0; i < n; i++)
    {
        clause_t *clause = &state->clauseList.clauses[clauseIndices[i]];
        updateClause(state, clause, assignment, clauseIndices[i]);
    }
}

/**
 * Remove clause from state and update all other values correspondingly, i.e., from update and ready2fire list.
 *
 * For removing always similarlly pattern
 *
 * Pattern for deleting from lists
 *
 * int index2Remove = get the index from the clause
 * int_list_t *list = getTheList
 * list->n--;
 * if (index2Remove != list->n)
 * {
 *      list->elems[index2Remove] = list->elems[list->n]
 *      clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
 *      cl->fieldToUpdate = index2Remove;
 * }
 *
 *
 * Pattern for shifted clause with newIndex
 *
 * int_list_t *list = getTheList
 * list[shiftedClause->ABC] = newIndex;
 */
void removeClause2(clauses_state_t *state, int index)
{
    if (DEBUG)
        printf("Remove clause\n");
    clause_t clause = state->clauseList.clauses[index];

    // delete info from additional lists
    {
        int index2Remove = clause.indexB1;
        int_list_t *list = &(state->baseLits2clauses[clause.b1]);
        list->n--;
        if (index2Remove != list->n)
        {
            list->elems[index2Remove] = list->elems[list->n];
            clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
            if (cl->b1 == clause.b1) // some special care needed
                cl->indexB1 = index2Remove;
            else
                cl->indexB2 = index2Remove;
        }
    }
    {
        int index2Remove = clause.indexB2;
        int_list_t *list = &(state->baseLits2clauses[clause.b2]);
        list->n--;
        if (index2Remove != list->n)
        {
            list->elems[index2Remove] = list->elems[list->n];
            clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
            if (cl->b2 == clause.b2) // some special care needed
                cl->indexB2 = index2Remove;
            else
                cl->indexB1 = index2Remove;
        }
    }

    removeP1andP2(state, &clause);

    // delete from ready to fire if present
    if (clause.indexFire != -1)
        removeReady2fire(state, &clause);

    state->clauseList.n--;
    if (index != state->clauseList.n) // put last element from list to the index and update information
    {
        // Update the indices according to shift
        state->clauseList.clauses[index] = state->clauseList.clauses[state->clauseList.n];
        clause_t *shiftedClause = &(state->clauseList.clauses[index]);

        for (int i = 0; i < shiftedClause->n; i++)
        {
            if (shiftedClause->indices[i] != -1)
            {
                lit_t l = shiftedClause->literals[i];
                state->lit2clauses[l].elems[shiftedClause->indices[i]] = index;
            }
        }

        if (shiftedClause->indexFire != -1)
        {
            state->ready2fire.elems[shiftedClause->indexFire] = index;
        }

        state->baseLits2clauses[shiftedClause->b1].elems[shiftedClause->indexB1] = index;
        state->baseLits2clauses[shiftedClause->b2].elems[shiftedClause->indexB2] = index;
    }
}

// -----------------------------------------helper functions--------------------------------------

/**
 *  For debugging. Check if clause_state is conform with real assignment.
 *  Does not check if there are duplicated clauses (i.e., clauses with identical literals), or clauses missing.
 */
void checkState2(clauses_state_t *state, truth_value_t *assignment, propagator_t *data)
{
    if (DEBUG)
        printf("c\tCheck %d clauses\n", state->clauseList.n);

    // check each clause
    for (int i = 0; i < state->clauseList.n; i++)
    {
        clause_t cl = state->clauseList.clauses[i];

        if (state->baseLits2clauses[cl.b1].elems[cl.indexB1] != i)
            EXIT_UNWANTED_STATE

        if (state->baseLits2clauses[cl.b2].elems[cl.indexB2] != i)
            EXIT_UNWANTED_STATE

        for (int j = 0; j < cl.n; j++)
        {
            if (cl.indices[j] != -1)
            {
                if (state->lit2clauses[cl.literals[j]].elems[cl.indices[j]] != i)
                    EXIT_UNWANTED_STATE
            }
        }

        if (cl.indexFire != -1)
        {
            if (state->ready2fire.elems[cl.indexFire] != i)
            {
                printf("%d %d \n", state->ready2fire.elems[cl.indexFire], i);
                EXIT_UNWANTED_STATE
            }
        }

        if (cl.p1 != -1 && assignment[cl.literals[cl.p1]] == truth_value_false)
            EXIT_UNWANTED_STATE

        if (cl.p2 != -1 && assignment[cl.literals[cl.p2]] == truth_value_false)
            EXIT_UNWANTED_STATE

        if (cl.p1 != -1 && cl.indices[cl.p1] == -1)
        {
            printf("%d %d\n", cl.p1, cl.indices[cl.p1]);
            EXIT_UNWANTED_STATE
        }

        if (cl.p2 != -1 && cl.indices[cl.p2] == -1)
            EXIT_UNWANTED_STATE

        if (cl.p1 > cl.p2 && cl.p2 != -1)
            EXIT_UNWANTED_STATE

        // if unit or unsatisfied then it must be in ready to fire list !!!!!!!!!!! most important thing
        int nFalse = 0;
        int nTrue = 0;
        for (int j = 0; j < cl.n; j++)
        {
            lit_t lit = cl.literals[j];
            if (assignment[lit] == truth_value_false)
                nFalse++;
            if (assignment[lit] == truth_value_true)
                nTrue++;
        }

        if (nFalse >= cl.n - 1 && nTrue == 0 && cl.indexFire == -1)
        {
            printf("%d %d %d %d\n", nFalse, nTrue, cl.n, cl.indexFire);
            EXIT_UNWANTED_STATE
        }
    }

    // go over baseLits2clauses lists
    for (int i = 0; i < data->inst->solveLitMax; i++)
    {
        if (assignment[i] != truth_value_true) // not allowed to have clauses
        {
            if (state->baseLits2clauses[i].n != 0)
                EXIT_UNWANTED_STATE;
        }
        else
        {
            for (int j = 0; j < state->baseLits2clauses[i].n; j++)
            {
                if (state->baseLits2clauses[i].elems[j] >= state->clauseList.n)
                {
                    printf("erorlit: %d; %d >= %d\n", i, state->baseLits2clauses[i].elems[j], state->clauseList.n);
                    EXIT_UNWANTED_STATE
                }
                clause_t cl = state->clauseList.clauses[state->baseLits2clauses[i].elems[j]];

                if (cl.b1 != i && cl.b2 != i)
                {
                    printf("b1: %d, b2: %d, index: %d\n", cl.b1, cl.b2, i);

                    EXIT_UNWANTED_STATE
                }
            }
        }
    }

    // go over update lists
    for (int i = 0; i < data->inst->solveLitMax; i++)
    {
        for (int j = 0; j < state->lit2clauses[i].n; j++)
        {
            int index = state->lit2clauses[i].elems[j];
            if (index >= state->clauseList.n)
                EXIT_UNWANTED_STATE
            clause_t cl = state->clauseList.clauses[index];

            // atleast one of the literals must be the clause
            bool literalPresent = false;
            for (int k = 0; k < cl.n; k++)
            {
                if (cl.literals[k] == i)
                {
                    literalPresent = true;
                    if (cl.indices[k] != j)
                    {
                        printf("%d %d\n", cl.indices[k], j);
                        printf("p1: %d, p2: %d, t1: %d, t2: %d\n", cl.p1, cl.p2, assignment[cl.literals[cl.p1]], assignment[cl.literals[cl.p2]]);
                        EXIT_UNWANTED_STATE
                    }
                    break;
                }
            }
            if (!literalPresent)
            {
                printf("Lit %d, clause: ", i);
                for (int k = 0; k < cl.n; k++)
                {
                    printf("%d ", cl.literals[k]);
                }
                printf("\n");
                EXIT_UNWANTED_STATE
            }
        }
    }

    for (int j = 0; j < state->ready2fire.n; j++)
    {
        int index = state->ready2fire.elems[j];
        if (index >= state->clauseList.n)
            EXIT_UNWANTED_STATE
        clause_t cl = state->clauseList.clauses[index];

        if (cl.indexFire != j)
            EXIT_UNWANTED_STATE
    }

    if (DEBUG)
        printf("Check completed\n");
}
