#include <clingo.h>
#include "util.h"
#include "args.h"
#include "main.h"
#include "solver.h"

my_args_t arguments;

int main(int argc, char const **argv)
{
    memset(&arguments, 0, sizeof(my_args_t));
    parse(argv, argc, &arguments);
    init_static();

    rotta_inst_t *solver = malloc(sizeof(rotta_inst_t));
    CHECK_OUT_OF_MEMORY(solver);
    memset(solver, 0, sizeof(rotta_inst_t));

    init(solver);

    solver->arguments.frequency = arguments.frequency;
    solver->minTransversals = arguments.minTransversals;

    if (!arguments.assumptionFile)
    {
        inst_solving_t solveData = {0};
        solveData.timeout = arguments.timeout;

        solve(solver, &solveData);
        if (solveData.finished)
        {
            // Currently only prints states for first thread
            printf("c\tSolving finished, Time %f\n", ((double)solveData.time) / CLOCKS_PER_SEC);
            printf("c\tClauses added: %ld\n", solver->propData->states[0].addedClauses);
            printf("c\tTime for base exchange: %ld\n", solver->propData->states[0].propTime / CLOCKS_PER_SEC);
            printf("c\tTime for maximility check: %ld\n", solver->propData->states[0].maximilityCheckTime / CLOCKS_PER_SEC);
        }
        else
        {
            printf("c\tNot finished\n");
        }
    }
    else
    {
        FILE *f = fopen(arguments.assumptionFile, "r");
        char *line = NULL;
        size_t n = 0;

        if (arguments.assumptionLine != 0)
        {
            for (int i = 0; i < arguments.assumptionLine - 1; i++)
                getline(&line, &n, f);
        }

        while (getline(&line, &n, f) != -1)
        {
            inst_solving_t solveData = {0};
            solveData.timeout = arguments.timeout;

            if (strncmp(line, "c\t", 2) == 0) // skip comments
                continue;

            printf("%s", line);

            char *s = strtok(line, " \n");
            while (s != NULL)
            {
                int tuple[RANK];

                int neg = 0;
                if (s[0] == '-')
                {
                    neg = 1;
                    s++;
                }
                s++;
                char *ep = s;

                for (int i = 0; i < RANK; i++)
                {
                    s = ep;
                    s++;
                    tuple[i] = (int)strtol(s, &ep, 10);
                }

                if (!neg)
                {
                    for (int i = 0; i < RANK; i++)
                        solveData.presentTuples[solveData.nPresentTuples][i] = tuple[i];
                    solveData.nPresentTuples++;
                }
                else
                {
                    for (int i = 0; i < RANK; i++)
                        solveData.notPresentTuples[solveData.nNotPresentTuples][i] = tuple[i];
                    solveData.nNotPresentTuples++;
                }

                s = strtok(NULL, " \n");
            }

            solve(solver, &solveData);

            if (solveData.finished)
            {
                printf("c\tSolving finished in %f\n", ((double)solveData.time) / CLOCKS_PER_SEC);
            }
            else
            {
                printf("c\tNot finished\n");
            }
            return EXIT_SUCCESS;
        }
        // TODO check if indeed end of file error
    }

    return EXIT_SUCCESS;
}